@@ -1,5 +1,5 @@
 /* GNU m4 -- A simple macro processor
-   Copyright (C) 1989, 90, 91, 92, 93 Free Software Foundation, Inc.
+   Copyright (C) 1989, 90, 91, 92, 93, 94 Free Software Foundation, Inc.
   
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,9 +20,13 @@
    expansion of user defined macros.  */
 
 #include "m4.h"
+
+extern FILE *popen ();
+
 #ifndef size_t
 #include <sys/types.h>
 #endif
+
 #include "regex.h"
 
 #define ARG(i)	(argc > (i) ? TOKEN_DATA_TEXT (argv[i]) : "")
@@ -42,6 +46,9 @@ DECLARE (m4___line__);
 DECLARE (m4_builtin);
 DECLARE (m4_changecom);
 DECLARE (m4_changequote);
+#ifdef ENABLE_CHANGEWORD
+DECLARE (m4_changeword);
+#endif
 DECLARE (m4_debugmode);
 DECLARE (m4_debugfile);
 DECLARE (m4_decr);
@@ -86,13 +93,16 @@ static builtin
 builtin_tab[] =
 {
 
-  /* name		gnu	args	blind	function */
+  /* name		GNUext	macros	blind	function */
 
   { "__file__",		TRUE,	FALSE,	FALSE,	m4___file__ },
   { "__line__",		TRUE,	FALSE,	FALSE,	m4___line__ },
   { "builtin",		TRUE,	FALSE,	TRUE,	m4_builtin },
   { "changecom",	FALSE,	FALSE,	FALSE,	m4_changecom },
   { "changequote",	FALSE,	FALSE,	FALSE,	m4_changequote },
+#ifdef ENABLE_CHANGEWORD
+  { "changeword",	TRUE,	FALSE,	FALSE,	m4_changeword },
+#endif
   { "debugmode",	TRUE,	FALSE,	FALSE,	m4_debugmode },
   { "debugfile",	TRUE,	FALSE,	FALSE,	m4_debugfile },
   { "decr",		FALSE,	FALSE,	TRUE,	m4_decr },
@@ -158,15 +168,13 @@ prefix_all_builtins (void)
 
   char *string;
   builtin *cursor;
-  builtin *limit;
 
   if (already_prefixed)
     return;
 
-  limit = builtin_tab + (sizeof (builtin_tab) / sizeof (builtin));
-  for (cursor = builtin_tab; cursor < limit; cursor++)
+  for (cursor = builtin_tab; cursor->name != NULL; cursor++)
     {
-      string = xmalloc (strlen (cursor->name) + 4);
+      string = (char *) xmalloc (strlen (cursor->name) + 4);
       strcpy (string, "m4_");
       strcat (string, cursor->name);
       cursor->name = string;
@@ -194,8 +202,8 @@ find_builtin_by_addr (builtin_func *func)
 | Find the builtin, which has NAME.  |
 `-----------------------------------*/
 
-const builtin *
-find_builtin_by_name (char *name)
+static const builtin *
+find_builtin_by_name (const char *name)
 {
   const builtin *bp;
 
@@ -294,12 +302,15 @@ bad_argc (token_data *name, int argc, int min, int max)
   if (min > 0 && argc < min)
     {
       if (!suppress_warnings)
-	warning ("too few arguments to built-in `%s'", TOKEN_DATA_TEXT (name));
+	M4ERROR ((warning_status, 0,
+		  "Warning: Too few arguments to built-in `%s'",
+		  TOKEN_DATA_TEXT (name)));
       isbad = TRUE;
     }
   else if (max > 0 && argc > max && !suppress_warnings)
-    warning ("excess arguments to built-in `%s' ignored",
-	     TOKEN_DATA_TEXT (name));
+    M4ERROR ((warning_status, 0,
+	      "Warning: Excess arguments to built-in `%s' ignored",
+	      TOKEN_DATA_TEXT (name)));
 
   return isbad;
 }
@@ -311,11 +322,15 @@ bad_argc (token_data *name, int argc, int min, int max)
 `--------------------------------------------------------------------------*/
 
 static boolean
-numeric_arg (token_data *macro, char *arg, int *valuep)
+numeric_arg (token_data *macro, const char *arg, int *valuep)
 {
-  if (sscanf (arg, "%d", valuep) != 1)
+  char *endp;
+
+  if (*arg == 0 || (*valuep = strtol (arg, &endp, 10), *endp != 0))
     {
-      m4error ("non-numeric argument to built-in `%s'", TOKEN_DATA_TEXT (macro));
+      M4ERROR ((warning_status, 0,
+		"Non-numeric argument to built-in `%s'",
+		TOKEN_DATA_TEXT (macro)));
       return FALSE;
     }
   return TRUE;
@@ -329,11 +344,11 @@ numeric_arg (token_data *macro, char *arg, int *valuep)
 /* Digits for number to ascii conversions.  */
 static char const digits[] = "0123456789abcdefghijklmnopqrstuvwxyz";
 
-static char *
-ntoa (register int value, int radix)
+static const char *
+ntoa (register eval_t value, int radix)
 {
   boolean negative;
-  unsigned int uvalue;
+  unsigned_eval_t uvalue;
   static char str[256];
   register char *s = &str[sizeof str];
 
@@ -342,12 +357,12 @@ ntoa (register int value, int radix)
   if (value < 0)
     {
       negative = TRUE;
-      uvalue = (unsigned int) -value;
+      uvalue = (unsigned_eval_t) -value;
     }
   else
     {
       negative = FALSE;
-      uvalue = (unsigned int) value;
+      uvalue = (unsigned_eval_t) value;
     }
 
   do
@@ -370,9 +385,9 @@ ntoa (register int value, int radix)
 static void
 shipout_int (struct obstack *obs, int val)
 {
-  char *s;
+  const char *s;
 
-  s = ntoa (val, 10);
+  s = ntoa ((eval_t) val, 10);
   obstack_grow (obs, s, strlen (s));
 }
 
@@ -383,7 +398,7 @@ shipout_int (struct obstack *obs, int val)
 
 static void
 dump_args (struct obstack *obs, int argc, token_data **argv,
-	   char *sep, boolean quoted)
+	   const char *sep, boolean quoted)
 {
   int i;
   int len = strlen (sep);
@@ -399,7 +414,6 @@ dump_args (struct obstack *obs, int argc, token_data **argv,
       if (quoted)
 	obstack_grow (obs, rquote, len_rquote);
     }
-
 }
 
 /* The rest of this file is code for builtins and expansion of user
@@ -444,6 +458,7 @@ define_macro (int argc, token_data **argv, symbol_lookup mode)
     case TOKEN_TEXT:
       define_user_macro (ARG (1), ARG (2), mode);
       break;
+
     case TOKEN_FUNC:
       bp = find_builtin_by_addr (TOKEN_DATA_FUNC (argv[2]));
       if (bp == NULL)
@@ -451,9 +466,11 @@ define_macro (int argc, token_data **argv, symbol_lookup mode)
       else
 	define_builtin (ARG (1), bp, mode, TOKEN_DATA_FUNC_TRACED (argv[2]));
       break;
+
     default:
-      internal_error ("Bad token data type in define_macro ()");
-      break;
+      M4ERROR ((warning_status, 0,
+		"INTERNAL ERROR: Bad token data type in define_macro ()"));
+      abort ();
     }
   return;
 }
@@ -495,7 +512,7 @@ static void
 m4_ifdef (struct obstack *obs, int argc, token_data **argv)
 {
   symbol *s;
-  char *result;
+  const char *result;
 
   if (bad_argc (argv[0], argc, 3, 4))
     return;
@@ -515,7 +532,7 @@ m4_ifdef (struct obstack *obs, int argc, token_data **argv)
 static void
 m4_ifelse (struct obstack *obs, int argc, token_data **argv)
 {
-  char *result;
+  const char *result;
   token_data *argv0;
 
   if (argc == 2)
@@ -588,7 +605,7 @@ dump_symbol (symbol *sym, struct dump_symbol_data *data)
 `------------------------------------------------------------------------*/
 
 static int
-dumpdef_cmp (const void *s1, const void *s2)
+dumpdef_cmp (const voidstar s1, const voidstar s2)
 {
   return strcmp (SYMBOL_NAME (* (const symbol **) s1),
 		 SYMBOL_NAME (* (const symbol **) s2));
@@ -623,7 +640,8 @@ m4_dumpdef (struct obstack *obs, int argc, token_data **argv)
 	  if (s != NULL && SYMBOL_TYPE (s) != TOKEN_VOID)
 	    dump_symbol (s, &data);
 	  else
-	    m4error ("Undefined name %s", TOKEN_DATA_TEXT (argv[i]));
+	    M4ERROR ((warning_status, 0,
+		      "Undefined name %s", TOKEN_DATA_TEXT (argv[i])));
 	}
     }
 
@@ -635,24 +653,33 @@ m4_dumpdef (struct obstack *obs, int argc, token_data **argv)
 
   for (; data.size > 0; --data.size, data.base++)
     {
-      debug_print ("%s:\t", SYMBOL_NAME (data.base[0]));
+      DEBUG_PRINT1 ("%s:\t", SYMBOL_NAME (data.base[0]));
 
       switch (SYMBOL_TYPE (data.base[0]))
 	{
 	case TOKEN_TEXT:
 	  if (debug_level & DEBUG_TRACE_QUOTE)
-	    debug_print ("%s%s%s\n", lquote, SYMBOL_TEXT (data.base[0]), rquote);
+	    DEBUG_PRINT3 ("%s%s%s\n",
+			  lquote, SYMBOL_TEXT (data.base[0]), rquote);
 	  else
-	    debug_print ("%s\n", SYMBOL_TEXT (data.base[0]));
+	    DEBUG_PRINT1 ("%s\n", SYMBOL_TEXT (data.base[0]));
 	  break;
+
 	case TOKEN_FUNC:
 	  bp = find_builtin_by_addr (SYMBOL_FUNC (data.base[0]));
 	  if (bp == NULL)
-	    internal_error ("built-in not found in builtin table!");
-	  debug_print ("<%s>\n", bp->name);
+	    {
+	      M4ERROR ((warning_status, 0, "\
+INTERNAL ERROR: Built-in not found in builtin table!"));
+	      abort ();
+	    }
+	  DEBUG_PRINT1 ("<%s>\n", bp->name);
 	  break;
+
 	default:
-	  internal_error ("Bad token data type in m4_dumpdef ()");
+	  M4ERROR ((warning_status, 0,
+		    "INTERNAL ERROR: Bad token data type in m4_dumpdef ()"));
+	  abort ();
 	  break;
 	}
     }
@@ -669,14 +696,15 @@ static void
 m4_builtin (struct obstack *obs, int argc, token_data **argv)
 {
   const builtin *bp;
-  char *name = ARG (1);
+  const char *name = ARG (1);
 
   if (bad_argc (argv[0], argc, 2, -1))
     return;
 
   bp = find_builtin_by_name (name);
   if (bp == NULL)
-    m4error ("Undefined name %s", name);
+    M4ERROR ((warning_status, 0,
+	      "Undefined name %s", name));
   else
     (*bp->func) (obs, argc - 1, argv + 1);
 }
@@ -692,14 +720,15 @@ static void
 m4_indir (struct obstack *obs, int argc, token_data **argv)
 {
   symbol *s;
-  char *name = ARG (1);
+  const char *name = ARG (1);
 
   if (bad_argc (argv[0], argc, 1, -1))
     return;
 
   s = lookup_symbol (name, SYMBOL_LOOKUP);
   if (s == NULL)
-    m4error ("Undefined macro `%s'", name);
+    M4ERROR ((warning_status, 0,
+	      "Undefined macro `%s'", name));
   else
     call_macro (s, argc - 1, argv + 1, obs);
 }
@@ -729,14 +758,18 @@ m4_defn (struct obstack *obs, int argc, token_data **argv)
       obstack_grow (obs, SYMBOL_TEXT (s), strlen (SYMBOL_TEXT (s)));
       obstack_grow (obs, rquote, len_rquote);
       break;
+
     case TOKEN_FUNC:
       push_macro (SYMBOL_FUNC (s), SYMBOL_TRACED (s));
       break;
+
     case TOKEN_VOID:
       break;
+
     default:
-      internal_error ("Bad symbol type in m4_defn ()");
-      break;
+      M4ERROR ((warning_status, 0,
+		"INTERNAL ERROR: Bad symbol type in m4_defn ()"));
+      abort ();
     }
 }
 
@@ -754,6 +787,8 @@ m4_syscmd (struct obstack *obs, int argc, token_data **argv)
 {
   if (bad_argc (argv[0], argc, 2, 2))
     return;
+
+  debug_flush_files ();
   sysval = system (ARG (1));
 }
 
@@ -766,10 +801,12 @@ m4_esyscmd (struct obstack *obs, int argc, token_data **argv)
   if (bad_argc (argv[0], argc, 2, 2))
     return;
 
+  debug_flush_files ();
   pin = popen (ARG (1), "r");
   if (pin == NULL)
     {
-      m4error ("Cannot open pipe to command \"%s\": %s", ARG (1), syserr ());
+      M4ERROR ((warning_status, errno,
+		"Cannot open pipe to command \"%s\"", ARG (1)));
       sysval = 0xff << 8;
     }
   else
@@ -795,10 +832,10 @@ m4_sysval (struct obstack *obs, int argc, token_data **argv)
 static void
 m4_eval (struct obstack *obs, int argc, token_data **argv)
 {
-  int value;
+  eval_t value;
   int radix = 10;
   int min = 1;
-  char *s;
+  const char *s;
 
   if (bad_argc (argv[0], argc, 2, 4))
     return;
@@ -806,9 +843,10 @@ m4_eval (struct obstack *obs, int argc, token_data **argv)
   if (argc >= 3 && !numeric_arg (argv[0], ARG (2), &radix))
     return;
 
-  if (radix <= 1 || radix > strlen (digits))
+  if (radix <= 1 || radix > (int) strlen (digits))
     {
-      m4error ("radix in eval out of range (radix = %d)", radix);
+      M4ERROR ((warning_status, 0,
+		"Radix in eval out of range (radix = %d)", radix));
       return;
     }
 
@@ -816,7 +854,8 @@ m4_eval (struct obstack *obs, int argc, token_data **argv)
     return;
   if  (min <= 0)
     {
-      m4error ("negative width to eval");
+      M4ERROR ((warning_status, 0,
+		"Negative width to eval"));
       return;
     }
 
@@ -923,7 +962,8 @@ m4_undivert (struct obstack *obs, int argc, token_data **argv)
 	if (sscanf (ARG (i), "%d", &div) == 1)
 	  insert_diversion (div);
 	else if (no_gnu_extensions)
-	  m4error ("non-numeric argument to %s", TOKEN_DATA_TEXT (argv[0]));
+	  M4ERROR ((warning_status, 0,
+		    "Non-numeric argument to %s", TOKEN_DATA_TEXT (argv[0])));
 	else
 	  {
 	    fp = path_search (ARG (i));
@@ -933,14 +973,15 @@ m4_undivert (struct obstack *obs, int argc, token_data **argv)
 		fclose (fp);
 	      }
 	    else
-	      m4error ("can't undivert %s: %s", ARG (i), syserr ());
+	      M4ERROR ((warning_status, errno,
+			"Can't undivert %s", ARG (i)));
 	  }
       }
 }
 
 /* This section contains various macros, which does not fall into any
-   specific group.  These are "dnl", "shift", "changequote" and
-   "changecom".  */
+   specific group.  These are "dnl", "shift", "changequote", "changecom"
+   and "changeword".  */
 
 /*------------------------------------------------------------------------.
 | Delete all subsequent whitespace from input.  The function skip_line () |
@@ -998,6 +1039,25 @@ m4_changecom (struct obstack *obs, int argc, token_data **argv)
     set_comment (TOKEN_DATA_TEXT (argv[1]),
 		(argc >= 3) ? TOKEN_DATA_TEXT (argv[2]) : NULL);
 }
+
+#ifdef ENABLE_CHANGEWORD
+
+/*-----------------------------------------------------------------------.
+| Change the regular expression used for breaking the input into words.	 |
+| The function set_word_regexp () lives in input.c.			 |
+`-----------------------------------------------------------------------*/
+
+static void
+m4_changeword (struct obstack *obs, int argc, token_data **argv)
+{
+  if (bad_argc (argv[0], argc, 2, 2))
+    return;
+
+  set_word_regexp (TOKEN_DATA_TEXT (argv[1]));
+}
+
+#endif /* ENABLE_CHANGEWORD */
+
 
 /* This section contains macros for inclusion of other files -- "include"
    and "sinclude".  This differs from bringing back diversions, in that
@@ -1020,7 +1080,8 @@ include (int argc, token_data **argv, boolean silent)
   if (fp == NULL)
     {
       if (!silent)
-	m4error ("can't open %s: %s", ARG (1), syserr ());
+	M4ERROR ((warning_status, errno,
+		  "Can't open %s", ARG (1)));
       return;
     }
 
@@ -1144,7 +1205,7 @@ m4_m4wrap (struct obstack *obs, int argc, token_data **argv)
 `-----------------------------------------------------------------------*/
 
 static void
-set_trace (symbol *sym, char *data)
+set_trace (symbol *sym, const char *data)
 {
   SYMBOL_TRACED (sym) = (boolean) (data != NULL);
 }
@@ -1164,7 +1225,8 @@ m4_traceon (struct obstack *obs, int argc, token_data **argv)
 	if (s != NULL)
 	  set_trace (s, (char *) obs);
 	else
-	  m4error ("Undefined name %s", TOKEN_DATA_TEXT (argv[i]));
+	  M4ERROR ((warning_status, 0,
+		    "Undefined name %s", TOKEN_DATA_TEXT (argv[i])));
       }
 }
 
@@ -1187,7 +1249,8 @@ m4_traceoff (struct obstack *obs, int argc, token_data **argv)
 	if (s != NULL)
 	  set_trace (s, NULL);
 	else
-	  m4error ("Undefined name %s", TOKEN_DATA_TEXT (argv[i]));
+	  M4ERROR ((warning_status, 0,
+		    "Undefined name %s", TOKEN_DATA_TEXT (argv[i])));
       }
 }
 
@@ -1222,7 +1285,8 @@ m4_debugmode (struct obstack *obs, int argc, token_data **argv)
 	}
 
       if (new_debug_level < 0)
-	m4error ("debugmode: bad debug flags: `%s'", ARG (1));
+	M4ERROR ((warning_status, 0,
+		  "Debugmode: bad debug flags: `%s'", ARG (1)));
       else
 	{
 	  switch (change_flag)
@@ -1230,9 +1294,11 @@ m4_debugmode (struct obstack *obs, int argc, token_data **argv)
 	    case 0:
 	      debug_level = new_debug_level;
 	      break;
+
 	    case '+':
 	      debug_level |= new_debug_level;
 	      break;
+
 	    case '-':
 	      debug_level &= ~new_debug_level;
 	      break;
@@ -1255,7 +1321,8 @@ m4_debugfile (struct obstack *obs, int argc, token_data **argv)
   if (argc == 1)
     debug_set_output (NULL);
   else if (!debug_set_output (ARG (1)))
-    m4error ("cannot set error file: %s: %s", ARG (1), syserr ());
+    M4ERROR ((warning_status, errno,
+	      "Cannot set error file: %s", ARG (1)));
 }
 
 /* This section contains text processing macros: "len", "index",
@@ -1282,7 +1349,7 @@ m4_len (struct obstack *obs, int argc, token_data **argv)
 static void
 m4_index (struct obstack *obs, int argc, token_data **argv)
 {
-  char *cp, *last;
+  const char *cp, *last;
   int l1, l2, retval;
 
   if (bad_argc (argv[0], argc, 3, 3))
@@ -1343,8 +1410,8 @@ m4_substr (struct obstack *obs, int argc, token_data **argv)
 | is made backwards, thus 9-0 is the string 9876543210.			  |
 `------------------------------------------------------------------------*/
 
-static char *
-expand_ranges (char *s, struct obstack *obs)
+static const char *
+expand_ranges (const char *s, struct obstack *obs)
 {
   char from;
   char to;
@@ -1385,15 +1452,15 @@ expand_ranges (char *s, struct obstack *obs)
 static void
 m4_translit (struct obstack *obs, int argc, token_data **argv)
 {
-  register char *data, *tmp;
-  char *from, *to;
+  register const char *data, *tmp;
+  const char *from, *to;
   int tolen;
 
   if (bad_argc (argv[0], argc, 3, 4))
     return;
 
   from = ARG (2);
-  if (index (from, '-') != NULL)
+  if (strchr (from, '-') != NULL)
     {
       from = expand_ranges (from, obs);
       if (from == NULL)
@@ -1403,7 +1470,7 @@ m4_translit (struct obstack *obs, int argc, token_data **argv)
   if (argc == 4)
     {
       to = ARG (3);
-      if (index (to, '-') != NULL)
+      if (strchr (to, '-') != NULL)
 	{
 	  to = expand_ranges (to, obs);
 	  if (to == NULL)
@@ -1417,7 +1484,7 @@ m4_translit (struct obstack *obs, int argc, token_data **argv)
 
   for (data = ARG (1); *data; data++)
     {
-      tmp = (char *) index (from, *data);
+      tmp = strchr (from, *data);
       if (tmp == NULL)
 	{
 	  obstack_1grow (obs, *data);
@@ -1441,17 +1508,20 @@ m4_format (struct obstack *obs, int argc, token_data **argv)
   format (obs, argc - 1, argv + 1);
 }
 
-/*--------------------------------------------------------------------------.
-| Function to perform substitution by regular expressions.  Used by the	    |
-| builtins regexp and patsubst.  The changed text is place on the obstack.  |
-| The substitution is REPL, with \& substituted by VICTIM (of length	    |
-| LENGTH), and \N substituted by the text matched by the Nth parenthesized  |
-| sub-expression, taken from REGS.					    |
-`--------------------------------------------------------------------------*/
+/*-------------------------------------------------------------------------.
+| Function to perform substitution by regular expressions.  Used by the	   |
+| builtins regexp and patsubst.  The changed text is placed on the	   |
+| obstack.  The substitution is REPL, with \& substituted by this part of  |
+| VICTIM matched by the last whole regular expression, taken from REGS[0], |
+| and \N substituted by the text matched by the Nth parenthesized	   |
+| sub-expression, taken from REGS[N].					   |
+`-------------------------------------------------------------------------*/
+
+static int substitute_warned = 0;
 
 static void
-substitute (struct obstack *obs, char *victim, int length,
-	    char *repl, struct re_registers *regs)
+substitute (struct obstack *obs, const char *victim, const char *repl,
+	    struct re_registers *regs)
 {
   register unsigned int ch;
 
@@ -1466,16 +1536,28 @@ substitute (struct obstack *obs, char *victim, int length,
 
       switch ((ch = *repl++))
 	{
-	case '0': case '1': case '2': case '3': case '4':
-	case '5': case '6': case '7': case '8': case '9':
+	case '0':
+	  if (!substitute_warned)
+	    {
+	      M4ERROR ((warning_status, 0, "\
+WARNING: \\0 will disappear, use \\& instead in replacements"));
+	      substitute_warned = 1;
+	    }
+	  /* Fall through.  */
+
+	case '&':
+	  obstack_grow (obs, victim + regs->start[0],
+			regs->end[0] - regs->start[0]);
+	  break;
+
+	case '1': case '2': case '3': case '4': case '5': case '6':
+	case '7': case '8': case '9': 
 	  ch -= '0';
 	  if (regs->end[ch] > 0)
 	    obstack_grow (obs, victim + regs->start[ch],
 			  regs->end[ch] - regs->start[ch]);
 	  break;
-	case '&':
-	  obstack_grow (obs, victim, length);
-	  break;
+
 	default:
 	  obstack_1grow (obs, ch);
 	  break;
@@ -1493,9 +1575,9 @@ substitute (struct obstack *obs, char *victim, int length,
 static void
 m4_regexp (struct obstack *obs, int argc, token_data **argv)
 {
-  char *victim;			/* first argument */
-  char *regexp;			/* regular expression */
-  char *repl;			/* replacement string */
+  const char *victim;		/* first argument */
+  const char *regexp;		/* regular expression */
+  const char *repl;		/* replacement string */
 
   struct re_pattern_buffer buf;	/* compiled regular expression */
   struct re_registers regs;	/* for subexpression matches */
@@ -1508,8 +1590,6 @@ m4_regexp (struct obstack *obs, int argc, token_data **argv)
 
   victim = TOKEN_DATA_TEXT (argv[1]);
   regexp = TOKEN_DATA_TEXT (argv[2]);
-  if (argc == 4)
-    repl = TOKEN_DATA_TEXT (argv[3]);
 
   buf.buffer = NULL;
   buf.allocated = 0;
@@ -1519,7 +1599,8 @@ m4_regexp (struct obstack *obs, int argc, token_data **argv)
 
   if (msg != NULL)
     {
-      m4error ("bad regular expression: \"%s\": %s", regexp, msg);
+      M4ERROR ((warning_status, 0,
+		"Bad regular expression: `%s': %s", regexp, msg));
       return;
     }
 
@@ -1529,14 +1610,18 @@ m4_regexp (struct obstack *obs, int argc, token_data **argv)
 
   if (startpos  == -2)
     {
-      m4error ("error matching regular expression \"%s\"", regexp);
+      M4ERROR ((warning_status, 0,
+		"Error matching regular expression \"%s\"", regexp));
       return;
     }
 
   if (argc == 3)
     shipout_int (obs, startpos);
   else if (startpos >= 0)
-    substitute (obs, victim, length, repl, &regs);
+    {
+      repl = TOKEN_DATA_TEXT (argv[3]);
+      substitute (obs, victim, repl, &regs);
+    }
 
   return;
 }
@@ -1551,8 +1636,8 @@ m4_regexp (struct obstack *obs, int argc, token_data **argv)
 static void
 m4_patsubst (struct obstack *obs, int argc, token_data **argv)
 {
-  char *victim;			/* first argument */
-  char *regexp;			/* regular expression */
+  const char *victim;		/* first argument */
+  const char *regexp;		/* regular expression */
 
   struct re_pattern_buffer buf;	/* compiled regular expression */
   struct re_registers regs;	/* for subexpression matches */
@@ -1574,7 +1659,8 @@ m4_patsubst (struct obstack *obs, int argc, token_data **argv)
 
   if (msg != NULL)
     {
-      m4error ("bad regular expression: \"%s\": %s", regexp, msg);
+      M4ERROR ((warning_status, 0,
+		"Bad regular expression `%s': %s", regexp, msg));
       if (buf.buffer != NULL)
 	xfree (buf.buffer);
       return;
@@ -1591,33 +1677,32 @@ m4_patsubst (struct obstack *obs, int argc, token_data **argv)
 			    offset, length - offset, &regs);
       if (matchpos < 0)
 	{
-	  /*
-	   * Match failed -- either error or there is no match in the
-	   * rest of the string, in which case the rest of the string
-	   * is copied verbatim.
-	   */
+
+	  /* Match failed -- either error or there is no match in the
+	     rest of the string, in which case the rest of the string is
+	     copied verbatim.  */
+
 	  if (matchpos == -2)
-	    m4error ("error matching regular expression \"%s\"", regexp);
+	    M4ERROR ((warning_status, 0,
+		      "Error matching regular expression \"%s\"", regexp));
 	  else if (offset < length)
 	    obstack_grow (obs, victim + offset, length - offset);
 	  break;
 	}
-      /*
-       * Copy the part of the string that was skipped by re_search ().
-       */
+
+      /* Copy the part of the string that was skipped by re_search ().  */
+
       if (matchpos > offset)
 	obstack_grow (obs, victim + offset, matchpos - offset);
 
-      /*
-       * Handle the part of the string that was covered by the match.
-       */
-      substitute (obs, victim, length, ARG (3), &regs);
+      /* Handle the part of the string that was covered by the match.  */
+
+      substitute (obs, victim, ARG (3), &regs);
+
+      /* Update the offset to the end of the match.  If the regexp
+	 matched a null string, advance offset one more, to avoid
+	 infinite loops.  */
 
-      /*
-       * Update the offset to the end of the match.  If the regexp
-       * matched a null string, advance offset one more, to avoid
-       * infinite loops.
-       */
       offset = regs.end[0];
       if (regs.start[0] == regs.end[0])
 	obstack_1grow (obs, victim[offset++]);
@@ -1641,7 +1726,7 @@ void
 expand_user_macro (struct obstack *obs, symbol *sym,
 		   int argc, token_data **argv)
 {
-  register char *text;
+  register const char *text;
   int i;
 
   for (text = SYMBOL_TEXT (sym); *text != '\0';)
